#include <windows.h>
#include <stdio.h>
#include <tchar.h>
#include <Imagehlp.h>
#pragma warning (disable:4786 4511 4512 4100 4663 4710)
#pragma warning (push, 3)
#include <set>
#include <list>
#include <string>
using namespace std;
#pragma warning (pop)


#ifdef UNICODE 
#define tstring wstring
#else
#define tstring string
#endif

typedef set<tstring> STRING_SET;


class CFileBaseSize
{
  ULONGLONG m_pBase;
  ULONGLONG m_pEnd;
  tstring m_strFile;
public:
  CFileBaseSize(LPCTSTR szFile, ULONGLONG pBase,  ULONGLONG pEnd):
    m_pBase(pBase),m_pEnd(pEnd), m_strFile(szFile)
  {
    

  }
  const tstring & GetFileName(void)const{return m_strFile;}
  ULONGLONG GetBase(void)const{return m_pBase;}
  ULONGLONG GetEnd(void)const{return m_pEnd;}
};

typedef list<CFileBaseSize> FILEBASESIZE_LIST;
typedef FILEBASESIZE_LIST::const_iterator FBS_CITER;



/// 
/// local prototypes:
/// 
static BOOL ExamineDirectoryForDLLs(PBOOL pbResult, LPCTSTR szDir, LPCTSTR szWildcardExt, const STRING_SET &slExcludes, FILEBASESIZE_LIST fbsList, DWORD dwFlags);
static void Usage (int iExitCode);
static BOOL FillExcludeFileList(LPCTSTR szFile, STRING_SET &sl);


#define FLAG_MODE_VERBOSE 0x1

int _tmain(int argc, LPCTSTR argv[])
{
  if(argc<3)
    Usage(0);

  SetErrorMode(SEM_FAILCRITICALERRORS);
  STRING_SET slExcludes;
  FILEBASESIZE_LIST fbsList;

  DWORD dwAttr = GetFileAttributes(argv[1]);
  if(INVALID_FILE_ATTRIBUTES==dwAttr || !(dwAttr&FILE_ATTRIBUTE_DIRECTORY))
  {
    _tprintf(_T("Invalid directory name\n"));
    return 1;
  }
  int iIdx = 2;
  BOOL bResult = TRUE;
  DWORD dwFlags = 0L;

  if((argv[2][0]==_T('-') || argv[2][0]==_T('/')) )
  {

    if((argv[2][1]==_T('E')||argv[2][1]==_T('e')) && argv[2][2]==_T('=') &&  argv[2][3])
    {
      if(!FillExcludeFileList(&argv[2][3], slExcludes))
      {
        _tprintf(_T("Failed to read from %s\n"), &argv[2][3]);
        return 1;
      }
      iIdx++;
    }
    else if(argv[2][1]==_T('v')||argv[2][1]==_T('V'))
    {
      dwFlags |= FLAG_MODE_VERBOSE;
      iIdx++;
    }
  }

  for (;iIdx<argc;iIdx++)
  {

    if(!ExamineDirectoryForDLLs(&bResult, argv[1], argv[iIdx], slExcludes, fbsList, dwFlags))
    {
      _tprintf(_T("Examining %s for %s failed with %lu\n"), argv[1], argv[iIdx], GetLastError());
    }
  }

  _tprintf(_T("%sollisions detected\n"), bResult?_T("No c"):_T("C"));

	return !bResult;
}


void Usage (int iExitCode)
{
  _tprintf(_T("\nUsage:\nprefload dir <-e=excludefile> <-v> <wildcards> \n-v Verbose Mode\n\nprefload is Copyright (c) Stefan Kuhr 2006\n\n"));
  exit(iExitCode);
}


// MakePtr is a macro that allows you to easily add to values (including
// pointers) together without dealing with C's pointer arithmetic.  It
// essentially treats the last two parameters as DWORDs.  The first
// parameter is used to typecast the result to the appropriate pointer type.
#define MakePtr( cast, ptr, addValue ) (cast)( (DWORD_PTR)(ptr) + (DWORD_PTR)(addValue))

static BOOL GetBaseAndSize(LPCTSTR szWildCard, ULONGLONG *pBase, ULONGLONG *pSize)
{
  HANDLE hFile = INVALID_HANDLE_VALUE;
  HANDLE hFileMapping = NULL;
  PIMAGE_DOS_HEADER dosHeader = NULL;
  PIMAGE_FILE_HEADER pImgFileHdr = NULL;
  PIMAGE_NT_HEADERS pNTHeader = NULL;
	PIMAGE_NT_HEADERS64 pNTHeader64 = NULL;

  DWORD dwLastErr = ERROR_SUCCESS;
  PBYTE pMappedFileBase = NULL;
  hFile = CreateFile(szWildCard, GENERIC_READ, FILE_SHARE_READ, NULL,
                      OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
                  
  if ( hFile == INVALID_HANDLE_VALUE )
  {
      dwLastErr = GetLastError();
      _tprintf(_T("Couldn't open file with CreateFile with %lu\n"), dwLastErr);
      goto CLEANUP;
  }


  hFileMapping = CreateFileMapping(hFile, NULL, PAGE_READONLY, 0, 0, NULL);
  if (!hFileMapping)
  {
      dwLastErr = GetLastError();
      _tprintf(_T("Couldn't open file mapping for %s with %lu\n"), szWildCard, dwLastErr);
      goto CLEANUP;
  }

  pMappedFileBase = (PBYTE)MapViewOfFile(hFileMapping,FILE_MAP_READ,0,0,0);
  if (!pMappedFileBase)
  {
      dwLastErr = GetLastError();
      _tprintf(_T("Couldn't map view of file %s with %lu\n"), szWildCard, dwLastErr);
      goto CLEANUP;
  }

  dosHeader = (PIMAGE_DOS_HEADER)pMappedFileBase;
	pImgFileHdr = (PIMAGE_FILE_HEADER)pMappedFileBase;

  if ( dosHeader->e_magic != IMAGE_DOS_SIGNATURE )
  {
    dwLastErr = ERROR_BAD_FORMAT;
    _tprintf(_T("File is not an executable %lu\n"), dwLastErr);
    goto CLEANUP;
  }

	// Make pointers to 32 and 64 bit versions of the header.
  pNTHeader = MakePtr( PIMAGE_NT_HEADERS, dosHeader,
                                dosHeader->e_lfanew );

	pNTHeader64 = (PIMAGE_NT_HEADERS64)pNTHeader;


	if ( IsBadReadPtr( pNTHeader, sizeof(pNTHeader->Signature) ) )
	{
    dwLastErr = ERROR_BAD_FORMAT;
    _tprintf(_T("Not a Portable Executable binary, %lu\n"), dwLastErr);
    goto CLEANUP;
	}

  if ( pNTHeader->Signature != IMAGE_NT_SIGNATURE )
  {
    dwLastErr = ERROR_BAD_FORMAT;
    _tprintf(_T("Not a Portable Executable binary, %lu\n"), dwLastErr);
    goto CLEANUP;
  }

  if ( pNTHeader->OptionalHeader.Magic == IMAGE_NT_OPTIONAL_HDR64_MAGIC )
  {
    PIMAGE_OPTIONAL_HEADER64 pOpt = (PIMAGE_OPTIONAL_HEADER64)&pNTHeader64->OptionalHeader;
    *pBase = pOpt->ImageBase;
    *pSize = pOpt->SizeOfImage;
  }
  else
  {
    /// 
    /// x86 style PE file
    /// 
    PIMAGE_OPTIONAL_HEADER32 pOpt = (PIMAGE_OPTIONAL_HEADER32)&pNTHeader64->OptionalHeader;
    *pBase = pOpt->ImageBase;
    *pSize = pOpt->SizeOfImage;
  }






CLEANUP:

  if(pMappedFileBase)
    UnmapViewOfFile(pMappedFileBase);


  if(hFileMapping)
    CloseHandle(hFileMapping);

  if(INVALID_HANDLE_VALUE != hFile)
    CloseHandle(hFile);


  if(ERROR_SUCCESS==dwLastErr)
    return TRUE;
  SetLastError(dwLastErr);
  return FALSE;
}



/// 
/// yeah, this is what the romans called "divide et impera":
/// 
void printULONGLONGAdress(ULONGLONG ul)
{
  if(ul>ULONG_MAX)
    _tprintf (_T("%.16x"), ul);
  else
    _tprintf (_T("%.8x"), (DWORD)ul);
}

void printAdresses(ULONGLONG ul1, ULONGLONG ul2)
{
  printULONGLONGAdress(ul1);
  _tprintf (_T("-"));
  printULONGLONGAdress(ul2);
}


#define STARDOTSTAR _T("*.*")

static BOOL ExamineDirectoryForDLLs(PBOOL pbResult, LPCTSTR szDir, LPCTSTR szWildcardExt, 
  const STRING_SET &slExcludes, FILEBASESIZE_LIST fbsList, DWORD dwFlags)
{
  WIN32_FIND_DATA wfd;
  ZeroMemory(&wfd, sizeof(WIN32_FIND_DATA));
  size_t sizeDir = _tcslen(szDir);
  size_t sizeWildcardExt = _tcslen(szWildcardExt);
  LPTSTR szWildCard = (LPTSTR)LocalAlloc(LPTR, (sizeDir+sizeWildcardExt+2 + MAX_PATH)*sizeof(TCHAR));
  if(!szWildCard)
  {
    DWORD dwLastErr = GetLastError();
    _tprintf(_T("ERROR: Allocation failed with %lu\n"), dwLastErr);
    return FALSE; //
  }

  memcpy(szWildCard, szDir, sizeDir*sizeof(TCHAR));
  if(szWildCard[sizeDir-1]!=_T('\\'))
  {
    szWildCard[sizeDir] = _T('\\');
    sizeDir++;
  }


  /// 
  /// first we recurse into subdirectories:
  /// 
  memcpy(&szWildCard[sizeDir], STARDOTSTAR, sizeof(STARDOTSTAR));
  HANDLE hFind = FindFirstFile(szWildCard, &wfd);

  
  if(INVALID_HANDLE_VALUE!=hFind)
  {
    do
    {
      if(wfd.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY && 
        _tcscmp(wfd.cFileName, _T(".")) && _tcscmp(wfd.cFileName, _T("..")) )
        {
          memcpy(&szWildCard[sizeDir], wfd.cFileName, MAX_PATH);    
          *pbResult &= ExamineDirectoryForDLLs(pbResult, szWildCard, szWildcardExt, 
            slExcludes, fbsList, dwFlags);
        }
    }
    while(FindNextFile(hFind, &wfd));

    FindClose(hFind);
  }

  memcpy(&szWildCard[sizeDir], STARDOTSTAR, sizeof(STARDOTSTAR));
  memcpy(&szWildCard[sizeDir], szWildcardExt, (sizeWildcardExt+1)*sizeof(TCHAR));

  hFind = FindFirstFile(szWildCard, &wfd);

  if(INVALID_HANDLE_VALUE!=hFind)
  {
    do
    {
      _tcslwr(wfd.cFileName);

      if(slExcludes.end()!=slExcludes.find(tstring(wfd.cFileName)))
        continue;

      memcpy(&szWildCard[sizeDir], wfd.cFileName, MAX_PATH);
      if(dwFlags&FLAG_MODE_VERBOSE)
        _tprintf(_T("Examining file %s\n"), szWildCard);
      ULONGLONG pBase = 0;
      ULONGLONG pEnd = 0;
      FBS_CITER citer;

      if(GetBaseAndSize(szWildCard, &pBase, &pEnd))
      {
        pEnd+=pBase;

        for(citer=fbsList.begin();citer!=fbsList.end();citer++)
        {
          if(pBase>=citer->GetBase() && pBase<citer->GetEnd() ||
            pEnd>=citer->GetBase() && pEnd<citer->GetEnd())
          {
            _tprintf (_T("%s ("), wfd.cFileName);
            printAdresses(pBase, pEnd);
            _tprintf (_T(") has a conflict with %s("), citer->GetFileName().c_str());
            printAdresses(citer->GetBase(), citer->GetEnd());
            _tprintf (_T(")\n"));
            *pbResult &= FALSE;
          }

        }
      }

      fbsList.insert(fbsList.end(), CFileBaseSize(wfd.cFileName, pBase, pEnd));
      
    }
    while(FindNextFile(hFind, &wfd));

    FindClose(hFind);
  }

  if(szWildCard)
    LocalFree(szWildCard);

  return TRUE;
}




static BOOL FillExcludeFileList(LPCTSTR szFile, STRING_SET &sl)
{
  FILE *fp = _tfopen(szFile, _T("r"));
  if(!fp)
  {
    
    return FALSE;
  }

  fseek(fp, 0, SEEK_END);
  long lFileSize = ftell(fp) + sizeof(TCHAR);
  rewind(fp);
  LPTSTR szLine = (LPTSTR)malloc(lFileSize);
  if(!szLine)
  {
    fclose(fp);
    return FALSE;
  }

  lFileSize/=sizeof(TCHAR);


  while(!feof(fp) && _fgetts(szLine, lFileSize, fp) && !ferror(fp))
  {
    size_t stLen = _tcslen(szLine);

    if(stLen)
    {
      if(szLine[stLen-1]==_T('\n'))
        szLine[stLen-1] = _T('\0');

      sl.insert(tstring(_tcslwr(szLine)));
    }
  }


  fclose(fp);
  free(szLine);
  return TRUE;
}

